# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Taneli Hukkinen
# Licensed to PSF under a Contributor Agreement.

__all__ = ("toml_tools",)

import os

# By changing this one line, we can run the tests against
# a different module name.
import toml_tools as toml_tools

def stem(file_path):
    #type(str) -> str
    return os.path.splitext(os.path.basename(file_path))[0]


# -*- coding: utf-8 -*-
from datetime import time

import pytest

from . import toml_tools

from toml_tools._re import timezone

def test_invalid_type_nested():
    with pytest.raises(TypeError):
        toml_tools.dumps({"bytearr": bytearray()})


def test_invalid_time():
    offset_time = time(23, 59, 59, tzinfo=timezone.utc)
    with pytest.raises(ValueError):
        toml_tools.dumps({"offset time": offset_time})



# -*- coding: utf-8 -*-

from . import toml_tools



def test_newline_before_table():
    actual = toml_tools.dumps({"table": {}})
    expected = "[table]\n"
    assert actual == expected

    actual = toml_tools.dumps({"table": {"nested": {}, "val3": 3}, "val2": 2, "val1": 1})
    expected = """\
val2 = 2
val1 = 1

[table]
val3 = 3

[table.nested]
"""
    assert actual == expected


def test_empty_doc():
    assert toml_tools.dumps({}) == ""


def test_dont_write_redundant_tables():
    actual = toml_tools.dumps({"tab1": {"tab2": {"tab3": {}}}})
    expected = "[tab1.tab2.tab3]\n"
    assert actual == expected


def test_multiline():
    multiline_string = (
        "This is longer than threshold!\n"
        "Should be formatted as a multiline basic string"
    )
    actual = toml_tools.dumps({"ml_string": multiline_string}, multiline_strings=True)
    expected = '''\
ml_string = """
This is longer than threshold!
Should be formatted as a multiline basic string"""
'''
    assert actual == expected


def test_only_tables():
    actual = toml_tools.dumps({"tab1": {}, "tab2": {}})
    expected = """\
[tab1]

[tab2]
"""
    assert actual == expected


def test_tricky_keys():
    actual = toml_tools.dumps({"f": 1, "tab1": {}, "": {"f": 2, "": {"": 1}}, "tab3": {}})
    expected = """\
f = 1

[tab1]

[""]
f = 2

["".""]
"" = 1

[tab3]
"""
    assert actual == expected


def test_nested_keys():
    actual = toml_tools.dumps(
        {
            "k": 1,
            "a": {"b": {"c": {"d": {"e": {"f": {}}, "e2": {"f2": {}}}, "d_key1": 1}}},
        }
    )
    expected = """\
k = 1

[a.b.c]
d_key1 = 1

[a.b.c.d.e.f]

[a.b.c.d.e2.f2]
"""
    assert actual == expected


def test_array_of_tables_containing_lists():
    example = {"aot": [{"a": [0, 1, 2, 3]}]}
    expected = """\
[[aot]]
a = [
    0,
    1,
    2,
    3,
]
"""
    actual = toml_tools.dumps(example)
    assert actual == expected
    assert toml_tools.loads(actual) == example

    example = {"a": {"nested": example}}
    expected = """\
[[a.nested.aot]]
a = [
    0,
    1,
    2,
    3,
]
"""
    actual = toml_tools.dumps(example)
    assert actual == expected


def test_array_of_long_tables():
    long_dict = {
        "long-value": "Lorem ipsum sith",
        "another-long-value": "consectetur adipis",
        "simple-value": 3,
    }
    example = {"table": {"nested-array": [{"a": 42}, long_dict]}}
    expected = """\
[[table.nested-array]]
a = 42

[[table.nested-array]]
long-value = "Lorem ipsum sith"
another-long-value = "consectetur adipis"
simple-value = 3
"""
    actual = toml_tools.dumps(example)
    assert actual == expected
    assert toml_tools.loads(actual) == example


def test_array_of_short_tables():
    long_name = "a" * 87
    example = {"table": {"nested-array": [{long_name: 0}, {"b": 1}, {"c": 2}]}}
    expected = """\
[table]
nested-array = [
    { %s = 0 },
    { b = 1 },
    { c = 2 },
]
""" % long_name
    actual = toml_tools.dumps(example)
    assert actual == expected


def test_example_issue_12():
    example = {
        "table": {
            "nested_table": [
                {"array_options": [1, 2, 3]},
                {"another_array": [1, 2]},
                {"c": 3},
            ]
        }
    }
    expected = """\
[[table.nested_table]]
array_options = [
    1,
    2,
    3,
]

[[table.nested_table]]
another_array = [
    1,
    2,
]

[[table.nested_table]]
c = 3
"""
    actual = toml_tools.dumps(example)
    assert actual == expected
    assert toml_tools.loads(actual) == example


def test_table_with_empty_array():
    # Empty arrays should never be AoTs
    example = {"table": {"array": []}}
    expected = """\
[table]
array = []
"""
    actual = toml_tools.dumps(example)
    assert actual == expected
    assert toml_tools.loads(actual) == example


def test_non_trivial_nesting():
    long = {
        "long-value": "Lorem ipsum dolor sit amet",
        "another-long-value": "consectetur adipiscing elit",
        "a-third-one": "sed do eiusmod tempor incididunt ut labore et dolore magna",
        "simple-value": 3,
    }
    example = {
        "table": {
            "aot": [
                {"nested-table": {"nested_aot": [{"a": [0, 1]}, {"b": 2}, {"c": 3}]}},
                {"other-nested-table": {"d": 4, "e": 5, "f": [{"g": 6}], "h": [long]}},
            ]
        }
    }

    expected = """\
[[table.aot]]

[[table.aot.nested-table.nested_aot]]
a = [
    0,
    1,
]

[[table.aot.nested-table.nested_aot]]
b = 2

[[table.aot.nested-table.nested_aot]]
c = 3

[[table.aot]]

[table.aot.other-nested-table]
d = 4
e = 5
f = [
    { g = 6 },
]

[[table.aot.other-nested-table.h]]
long-value = "Lorem ipsum dolor sit amet"
another-long-value = "consectetur adipiscing elit"
a-third-one = "sed do eiusmod tempor incididunt ut labore et dolore magna"
simple-value = 3
"""
    actual = toml_tools.dumps(example)
    assert actual == expected
    assert toml_tools.loads(actual) == example


def test_multiline_in_aot():
    data = {"aot": [{"multiline_string": "line1\nline2"}]}
    assert (
        toml_tools.dumps(data, multiline_strings=True)
        == '''\
[[aot]]
multiline_string = """
line1
line2"""
'''
    )
    assert (
        toml_tools.dumps(data, multiline_strings=False)
        == """\
aot = [
    { multiline_string = "line1\\nline2" },
]
"""
    )



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Taneli Hukkinen
# Licensed to PSF under a Contributor Agreement.

import os
import json
import glob 
import unittest

from . import burntsushi, toml_tools, stem


class MissingFile:
    def __init__(self, path):
        self.path = path


DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

VALID_FILES = glob.glob(os.path.join(DATA_DIR, "valid", "**/*.toml"))
assert VALID_FILES, "Valid TOML test files not found"

_expected_files = []
for p in VALID_FILES:
    json_path = os.path.splitext(p)[0] + ".json"
    try:
        with open(json_path, 'rb') as f:
            text = json.loads(f.read().decode())
    except FileNotFoundError:
        text = MissingFile(json_path)
    _expected_files.append(text)
VALID_FILES_EXPECTED = tuple(_expected_files)

INVALID_FILES = glob.glob(os.path.join(DATA_DIR, "invalid", "**/*.toml"))
assert INVALID_FILES, "Invalid TOML test files not found"


class TestData(unittest.TestCase):
    pass

def make_valid_test(valid, expected):
    def test_valid(self, valid = valid, expected  = expected):

        if isinstance(expected, MissingFile):
            # For a poor man's xfail, assert that this is one of the
            # test cases where expected data is known to be missing.
            assert stem(valid) in {
                "qa-array-inline-nested-1000",
                "qa-table-inline-nested-1000",
            }
            return
        

        with open(valid, 'rb') as f:
            toml_str = f.read().decode()
        actual = toml_tools.loads(toml_str)
        actual = burntsushi.convert(actual)
        expected = burntsushi.normalize(expected)
        self.assertEqual(actual, expected)
    return test_valid

for valid, expected in zip(VALID_FILES, VALID_FILES_EXPECTED):
    setattr(TestData, 
            'test_valid_%s' % stem(valid).replace('-','_'), 
            make_valid_test(valid, expected))



def make_invalid_test(invalid):
    def test_invalid(self, invalid = invalid):
        with open(invalid,'rb') as f:
            toml_bytes = f.read()
            try:
                toml_str = toml_bytes.decode('utf8')
            except UnicodeDecodeError:
                # Some BurntSushi tests are not valid UTF-8. Skip those.

                assert True # a poorer man's xfail
                return
            with self.assertRaises(toml_tools.TOMLDecodeError):
                toml_tools.loads(toml_str)
    return test_invalid

for invalid in INVALID_FILES:
    setattr(TestData,
            'test_invalid_%s' % stem(invalid).replace('-','_'), 
            make_invalid_test(invalid))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Taneli Hukkinen
# Licensed to PSF under a Contributor Agreement.


from collections import namedtuple
from datetime import date, datetime, time
from decimal import Decimal
import string

from ._helpers import ReadOnlyDict

ASCII_CTRL = frozenset(chr(i) for i in range(32)) | frozenset(chr(127))
ILLEGAL_BASIC_STR_CHARS = frozenset('"\\') | ASCII_CTRL - frozenset("\t")
BARE_KEY_CHARS = frozenset(string.ascii_letters + string.digits + "-_")
ARRAY_TYPES = (list, tuple)
ARRAY_INDENT = " " * 4
MAX_LINE_LENGTH = 100

COMPACT_ESCAPES = ReadOnlyDict({"\u0008": "\\b",  # backspace
                                "\u000A": "\\n",  # linefeed
                                "\u000C": "\\f",  # form feed
                                "\u000D": "\\r",  # carriage return
                                "\u0022": '\\"',  # quote
                                "\u005C": "\\\\",  # backslash
                               }
                              )


def dump(__obj, __fp, multiline_strings = False):
    #type(dict, BinaryIO, bool) -> None
    ctx = Context(multiline_strings, {})
    for chunk in gen_table_chunks(__obj, ctx, name=""):
        __fp.write(chunk.encode(encoding = 'utf8'))


def dumps(__obj,  multiline_strings = False):
    #type(dict, bool) -> str
    ctx = Context(multiline_strings, {})
    return "".join(gen_table_chunks(__obj, ctx, name=""))




Context = namedtuple('Context', ('allow_multiline', 'inline_table_cache'))


def gen_table_chunks(table, ctx, name, inside_aot = False):
    #type(Mapping, Context, str bool) -> Iterator[str]
    yielded = False
    literals = []
    tables = []  
    for k, v in table.items():
        if isinstance(v, dict):
            tables.append((k, v, False))
        elif is_aot(v) and not all(is_suitable_inline_table(t, ctx) for t in v):
            tables.extend((k, t, True) for t in v)
        else:
            literals.append((k, v))

    if inside_aot or name and (literals or not tables):
        yielded = True
        yield ("[[%s]]\n" if inside_aot else "[%s]\n") % name

    if literals:
        yielded = True
        for k, v in literals:
            yield "%s = %s\n" % (format_key_part(k), format_literal(v, ctx))

    for k, v, in_aot in tables:
        if yielded:
            yield "\n"
        else:
            yielded = True
        key_part = format_key_part(k)
        display_name = ("%s.%s" % (name, key_part)) if name else key_part
        for chunk in gen_table_chunks(v, ctx, name=display_name, inside_aot=in_aot):
            yield chunk


def format_literal(obj, ctx, nest_level= 0):
    #type(type[any], Context, int) -> str
    if isinstance(obj, bool):
        return "true" if obj else "false"
    if isinstance(obj, (int, float, date, datetime)):
        return str(obj)
    if isinstance(obj, Decimal):
        return format_decimal(obj)
    if isinstance(obj, time):
        if obj.tzinfo:
            raise ValueError("TOML does not support offset times")
        return str(obj)
    if isinstance(obj, str):
        return format_string(obj, allow_multiline=ctx.allow_multiline)
    if isinstance(obj, ARRAY_TYPES):
        return format_inline_array(obj, ctx, nest_level)
    if isinstance(obj, dict):
        return format_inline_table(obj, ctx)
    raise TypeError("Object of type %s is not TOML serializable" % type(obj))


def format_decimal(obj):
    #type(Decimal) -> str
    if obj.is_nan():
        return "nan"
    if obj == Decimal("inf"):
        return "inf"
    if obj == Decimal("-inf"):
        return "-inf"
    return str(obj)


def format_inline_table(obj, ctx):
    #type(type[any], Context) -> str

    # check cache first
    obj_id = id(obj)
    if obj_id in ctx.inline_table_cache:
        return ctx.inline_table_cache[obj_id]

    if not obj:
        rendered = "{}"
    else:
        rendered = (
            "{ "
            + ", ".join(
                ("%s = %s" % (format_key_part(k), format_literal(v, ctx)))
                for k, v in obj.items()
            )
            + " }"
        )
    ctx.inline_table_cache[obj_id] = rendered
    return rendered


def format_inline_array(obj, ctx, nest_level):
    #type(tuple | list, Context, int) -> str
    if not obj:
        return "[]"
    item_indent = ARRAY_INDENT * (1 + nest_level)
    closing_bracket_indent = ARRAY_INDENT * nest_level
    return (
        "[\n"
        + ",\n".join(
            item_indent + format_literal(item, ctx, nest_level=nest_level + 1)
            for item in obj
        )
        + ",\n%s]" % closing_bracket_indent
    )


def format_key_part(part):
    #type(str) -> str
    if part and BARE_KEY_CHARS.issuperset(part):
        return part
    return format_string(part, allow_multiline=False)


def format_string(s, allow_multiline):
    #type(str, bool) -> str
    do_multiline = allow_multiline and "\n" in s
    if do_multiline:
        result = '"""\n'
        s = s.replace("\r\n", "\n")
    else:
        result = '"'

    pos = seq_start = 0
    while True:
        try:
            char = s[pos]
        except IndexError:
            result += s[seq_start:pos]
            if do_multiline:
                return result + '"""'
            return result + '"'
        if char in ILLEGAL_BASIC_STR_CHARS:
            result += s[seq_start:pos]
            if char in COMPACT_ESCAPES:
                if do_multiline and char == "\n":
                    result += "\n"
                else:
                    result += COMPACT_ESCAPES[char]
            else:
                result += "\\u" + hex(ord(char))[2:].rjust(4, "0")
            seq_start = pos + 1
        pos += 1


def is_aot(obj):
    #type(type[Any]) -> bool
    """Decides if an object behaves as an array of tables (i.e. a nonempty list
    of dicts)."""
    return bool(
        isinstance(obj, ARRAY_TYPES) and obj and all(isinstance(v, dict) for v in obj)
    )


def is_suitable_inline_table(obj, ctx):
    #type(dict, Context) -> bool
    """Use heuristics to decide if the inline-style representation is a good
    choice for a given table."""
    rendered_inline = "%s%s," % (ARRAY_INDENT, format_inline_table(obj, ctx))
    return len(rendered_inline) <= MAX_LINE_LENGTH and "\n" not in rendered_inline



# -*- coding: utf-8 -*-

from decimal import Decimal

from . import toml_tools


def test_decimal():
    obj = {
        "decimal-0": Decimal(0),
        "decimal-pi": Decimal("3.14159"),
        "decimal-inf": Decimal("inf"),
        "decimal-minus-inf": Decimal("-inf"),
        "decimal-nan": Decimal("nan"),
    }
    assert (
        toml_tools.dumps(obj)
        == """\
decimal-0 = 0
decimal-pi = 3.14159
decimal-inf = inf
decimal-minus-inf = -inf
decimal-nan = nan
"""
    )


def test_tuple():
    obj = {"empty-tuple": (), "non-empty-tuple": (1, (2, 3))}
    assert (
        toml_tools.dumps(obj)
        == """\
empty-tuple = []
non-empty-tuple = [
    1,
    [
        2,
        3,
    ],
]
"""
    )



# -*- coding: utf-8 -*-

from . import toml_tools


def test_dump(tmp_path):
    toml_obj = {"testing": "test\ntest"}
    path = tmp_path / "test.toml"
    with open(path, "wb") as f:
        toml_tools.dump(toml_obj, f)
    assert path.read_bytes().decode() == 'testing = "test\\ntest"\n'



# -*- coding: utf-8 -*-
"""Test for profiling.

This test can be useful for profiling, as most of the execution time
will be spent writing TOML instead of managing pytest execution
environment. To get and read profiler results:
  - `tox -e profile`
  - `firefox .tox/prof/combined.svg`
"""
import os

from . import toml_tools


path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "benchmark", "data.toml")
with open(path, 'rb') as f:
    benchmark_toml = f.read().decode('utf8')
data = toml_tools.loads(benchmark_toml)


def test_for_profiler():
    # increase the count here to reduce the impact of
    # setting up pytest execution environment. Let's keep
    # the count low by default because this is part of the
    # standard test suite.
    iterations = int(os.environ.get("PROFILER_ITERATIONS", 1))
    for _ in range(iterations):
        toml_tools.dumps(data)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: MIT
# SPDX-FileCopyrightText: 2021 Taneli Hukkinen
# Licensed to PSF under a Contributor Agreement.

import unittest

from . import toml_tools


class TestError(unittest.TestCase):
    def test_line_and_col(self):
        with self.assertRaises(toml_tools.TOMLDecodeError) as exc_info:
            toml_tools.loads("val=.")
        self.assertEqual(str(exc_info.exception), "Invalid value (at line 1, column 5)")

        with self.assertRaises(toml_tools.TOMLDecodeError) as exc_info:
            toml_tools.loads(".")
        self.assertEqual(
            str(exc_info.exception), "Invalid statement (at line 1, column 1)"
        )

        with self.assertRaises(toml_tools.TOMLDecodeError) as exc_info:
            toml_tools.loads("\n\nval=.")
        self.assertEqual(str(exc_info.exception), "Invalid value (at line 3, column 5)")

        with self.assertRaises(toml_tools.TOMLDecodeError) as exc_info:
            toml_tools.loads("\n\n.")
        self.assertEqual(
            str(exc_info.exception), "Invalid statement (at line 3, column 1)"
        )

    def test_missing_value(self):
        with self.assertRaises(toml_tools.TOMLDecodeError) as exc_info:
            toml_tools.loads("\n\nfwfw=")
        self.assertEqual(str(exc_info.exception), "Invalid value (at end of document)")

    def test_invalid_char_quotes(self):
        with self.assertRaises(toml_tools.TOMLDecodeError) as exc_info:
            toml_tools.loads("v = '\n'")
        self.assertTrue(" '\\n' " in str(exc_info.exception))

    def test_module_name(self):
        self.assertEqual(toml_tools.TOMLDecodeError().__module__, toml_tools.__name__)

    def test_invalid_parse_float(self):
        def dict_returner(s):
            #type(str) -> dict
            return {}

        def list_returner(s):
            #type(str) -> list
            return []

        for invalid_parse_float in (dict_returner, list_returner):
            with self.assertRaises(ValueError) as exc_info:
                toml_tools.loads("f=0.1", parse_float=invalid_parse_float)
            self.assertEqual(
                str(exc_info.exception), "parse_float must not return dicts or lists"
            )



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import toml_tools

with open("README.md") as readme_file:
    readme_string = readme_file.read()

setup(
    name="toml_tools",
    version=toml_tools.__version__,
    description="Tomli and Tomli-W for Python 2 and Iron Python",
    author="Taneli Hukkinen",
    author_email="hukkin@users.noreply.github.com",
    url="https://github.com/JamesParrott/toml_tools",
    packages=['toml_tools'],
    license="MIT",
    long_description=readme_string,
    long_description_content_type = 'text/markdown',
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: IronPython',
    ]
)



# -*- coding: utf-8 -*-

import os
import glob
from decimal import Decimal
from math import isnan

import pytest
from . import toml_tools, stem


PARENT_DIR = os.path.dirname(__file__)
COMPLIANCE_DIR = os.path.join(PARENT_DIR, "data", "toml-lang-compliance", "valid")
EXTRAS_DIR = os.path.join(PARENT_DIR, "data", "extras", "valid")

VALID_FILES = (glob.glob(os.path.join(COMPLIANCE_DIR,"**/*.toml")) + 
               glob.glob(os.path.join(EXTRAS_DIR, "**/*.toml")))


@pytest.mark.parametrize(
    "valid",
    VALID_FILES#),
    #ids=[os.path.splitext(p)[0] for p in VALID_FILES],
    # ids=[stem(p) for p in VALID_FILES],
)
def test_valid(valid):
    if stem(valid) in {"qa-array-inline-nested-1000", "qa-table-inline-nested-1000"}:
        pytest.xfail("This much recursion is not supported")
    with open(valid,'rb') as f:
        original_str = f.read().decode()
    # original_str = valid.read_bytes().decode()
    original_data = toml_tools.loads(original_str)
    dump_str = toml_tools.dumps(original_data)
    after_dump_data = toml_tools.loads(dump_str)
    assert replace_nans(after_dump_data) == replace_nans(original_data)


NAN = object()


def replace_nans(cont):
    #type(Union[dict, list] -> Union[dict, list])
    """Replace NaNs with a sentinel object to fix the problem that NaN is not
    equal to another NaN."""
    for k, v in cont.items() if isinstance(cont, dict) else enumerate(cont):
        if isinstance(v, (float, Decimal)) and isnan(v):
            cont[k] = NAN
        elif isinstance(v, dict) or isinstance(v, list):
            cont[k] = replace_nans(cont[k])
    return cont


@pytest.mark.parametrize(
    "obj,expected_str,multiline_strings",
    [
        ({"cr-newline": "foo\rbar"}, 'cr-newline = "foo\\rbar"\n', True),
        ({"crlf-newline": "foo\r\nbar"}, 'crlf-newline = """\nfoo\nbar"""\n', True),
    ],
)
def test_obj_to_str_mapping(obj, expected_str, multiline_strings):
    assert toml_tools.dumps(obj, multiline_strings=multiline_strings) == expected_str



# -*- coding: utf-8 -*-

class ReadOnlyDict(object):
    def __init__(self, _dict):
        self._dict = _dict

    def __getitem__(self, key):
        return self._dict[key]
    
    def __contains__(self, key):
        return key in self._dict
    
    def keys(self):
        return self._dict.keys()
    
    def values(self):
        return self._dict.values()
    
    def items(self):
        return self._dict.items()